# Copyright © 2012-2018 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

"""
Constants for various XML namespaces
"""

didjvu = 'http://jwilk.net/software/didjvu#'

dc = 'http://purl.org/dc/elements/1.1/'
rdf = 'http://www.w3.org/1999/02/22-rdf-syntax-ns#'
tiff = 'http://ns.adobe.com/tiff/1.0/'
xmp = 'http://ns.adobe.com/xap/1.0/'
xmpmm = 'http://ns.adobe.com/xap/1.0/mm/'
